# File Name: bed.py
# Created By: ZW
# Created On: 2022-12-19
# Purpose: defines class for BED6 and BED12 style intervals inherited from
#  Intervals class for use in gobjects


# module imports
# ----------------------------------------------------------------------------
from .intervals import Interval
from dataclasses import dataclass
from typing import Union
from re import match

# function definitions
# ----------------------------------------------------------------------------

# function check_and_convert_nums() takes a string and converts it to 
# a floating point python number if it looks like an integer or float,
# otherwise, it returns the string back unchanged
def check_and_convert_nums(string):
    pattern = "^[+-]?((\d+(\.\d+)?)|(\.\d+))$"
    mth = match(pattern, string)
    return (float(mth.string) if mth is not None else string)


# class definitions
# ----------------------------------------------------------------------------

# class Bed6() - child class of Interval() which adds score and strand.
# * complies with the BED6 standard found on the UCSC file format standards
#   webpage: https://genome.ucsc.edu/FAQ/FAQformat.html#format1
# 
# * the class adds two attributes to the Interval base class:
#    1. interval score (score); typically 1-1000 -required, missing denoted by '.'-
#    2. feature strandness (strand) -required, missingness denoted by '.'-
@dataclass(eq=False, order=False)
class Bed6(Interval):
    score: float
    strand: str

    # define a post-init method to coerce ints and floats properly
    def __post_init__(self):
        self.chromStart = int(self.chromStart)
        self.chromEnd = int(self.chromEnd)
        self.zero_idx_start = self.chromStart  # always inclusive
        self.zero_idx_end = self.chromEnd - 1  # always inclusive
        self.score = check_and_convert_nums(str(self.score))

    # define a custom printout representation for the Bed6
    def __repr__(self):
        spec_bed6 = (
            f" {self.chrom} {self.chromStart} {self.chromEnd} {self.name}"
            f" {self.score} {self.strand} "
        )
        return repr(f"Bed6({spec_bed6})")
    

# class Bed12() - child class of Bed6() which adds several more fields.
# * complies with the BED12 standard found on the UCSC file format standards
#   webpage: https://genome.ucsc.edu/FAQ/FAQformat.html#format1
# 
# * the class adds six additional attributes to the Bed6 base class:
#    1. thickStart- The starting position at which the feature is drawn thickly (e.g. start codon)
#    2. thickEnd- The ending position at which the feature is drawn thickly (e.g. stop codon)
#    3. itemRgb- An RGB value of the form R,G,B (e.g. 255,0,0) to determine the display color
#    4. blockCount- The number of blocks (e.g. exons) in the BED line.
#    5. blockSizes- A comma-separated string of the block sizes.
#    6. blockStarts- A comma-separated string of block starts. positions should be relative to chromStart
@dataclass(eq=False, order=False)
class Bed12(Bed6):
    thickStart: int
    thickEnd: int
    itemRgb: str
    blockCount: int
    blockSizes: str
    blockStarts: str

    # define a custom printout representation for the Bed6
    def __repr__(self):
        spec_bed12 = (
            f" {self.chrom} {self.chromStart} {self.chromEnd} {self.name}"
            f" {self.score} {self.strand} "
            f" {self.thickStart} {self.thickEnd} {self.itemRgb}"
            f" {self.blockCount} {self.blockSizes} {self.blockStarts} "
        )
        return repr(f"Bed12({spec_bed12})")


# File Name: intervalmap.py
# Created By: ZW
# Created On: 2023-01-15
# Purpose: defines IntervalMap structure which allows efficient searching of 
#   contiguous bed-like or gtf-like intervals with a supplied query interval

# module imports
# ----------------------------------------------------------------------------
from collections import OrderedDict

# class definitions
# ----------------------------------------------------------------------------
# define a class IntervalMap which acts as a searchable hashtable
# like structure in which to store contiguous intervals for searching
# intersections against. each child node is either of type 
#
#           root
#            |
#       /    |     \
#    chr1  chr2  .. chrN
#    / |    / |      / |
#   A  B   C  D      E  F

class IntervalNode:
    def __init__ (self, chrom, intervals, isSorted=False):
        self.chrom = chrom
        self.intervals = intervals
        self.zero_idx_start = None
        self.zero_idx_end = None
        self._set_bounds(isSorted=isSorted)

    def _set_bounds(self, isSorted):
        if not isSorted: self.intervals.sort()
        self.zero_idx_start = self.intervals[0].zero_idx_start
        self.zero_idx_end = max([i.zero_idx_end for i in self.intervals])


class IntervalMap:
    def __init__ (self, interval_list, interval_list_sorted=False, subnodes=20):
        self.interval_list_sorted = interval_list_sorted
        self.subnodes = subnodes
        self._map = OrderedDict()
        self.create_map(interval_list, interval_list_sorted)

    def create_map(self, interval_list, isSorted):
        print(f"Building Interval Map for {len(interval_list)} contiguous objects..")
        ilist = sorted(interval_list) if not isSorted else interval_list
         # place the intervals into a chromosome level map
        for i in ilist:
            if i.chrom not in self._map.keys(): self._map[i.chrom] = [i]
            else: self._map[i.chrom].append(i)

        # for each chromosome, divide intervals into self.subnodes bins
        for chrom in self._map.keys():
            clist = self._map[chrom]
            intervals_per_node = int(len(clist) / self.subnodes)
            if intervals_per_node < 1: 
                self.subnodes = len(clist)
                intervals_per_node = 1
            bins = []
            for bin_num in range(self.subnodes):
                node = IntervalNode(chrom, clist[:intervals_per_node], isSorted=True)
                clist = clist[intervals_per_node:]
                bins.append(node)
            if len(clist) > 0:
                bins[-1].intervals.extend(clist)
                bins[-1]._set_bounds(isSorted=True)
            self._map[chrom] = bins

    # define external accessor function for the map
    def get_map(self): return self._map


# File Name: bedpe.py
# Created By: ZW
# Created On: 2022-01-10
# Purpose: defines class for bedpe style objects (identical to the Bedtools BEDPE spec),
#  as defined by standards at https://bedtools.readthedocs.io/en/latest/content/general-usage.html

# module imports
# ----------------------------------------------------------------------------
from .bed import Bed6
from dataclasses import dataclass
from operator import lt, gt
from re import split as re_split
from re import match


# function definitions
# ----------------------------------------------------------------------------

# function prep_chrom_compare() splits contig names that are a combo of
# strings and digits to use as sorting keys in sort functions and similar 
# processes. for example: 'chr1' -> ('chr', 1). this function can be 
# used in the sorted() function with the key=prep_chrom_comp argument
def prep_chrom_comp(chrom: str):
    def recode(substr: str):
        return int(substr) if substr.isdigit() else substr.lower()
    return [recode(sub) for sub in re_split('([0-9]+)',chrom)]

# function check_and_convert_nums() takes a string and converts it to 
# a floating point python number if it looks like an integer or float,
# otherwise, it returns the string back unchanged
def check_and_convert_nums(string):
    pattern = "^[+-]?((\d+(\.\d+)?)|(\.\d+))$"
    mth = match(pattern, string)
    return (float(mth.string) if mth is not None else string)

# class definitions
# ----------------------------------------------------------------------------

# class Bedpe() - base class for all Bedpe-type objects which can be used for 
# storing paired-end style objects or long-range interactions like those created
# by promoter-capture HiC. 
# * indexing of genomic intervals follow BED conventions. This means that
#   chromosome scaffolds begin at base 0. (This differs from gtf-syle so care
#   must be taken when intersecting the two object types) the end of the interval
#   is exclusive [chromStart, chromEnd), creating slightly different intersection
#   rules than bed-style objects. 
#
# * The Bedpe class contains nine attributes: 
#     1. chrom - chromosome of first interval
#     2. chromStart - chromosome start position for first interval [ 0-indexed, exclusive ]
#     3. chromEnd - chromosome end position for first interval [ 0-indexed, exclusive ]
#     4. chrom2 - chromosome of second interval
#     5. chromStart2 - chromosome start position for second interval [ 0-indexed, exclusive ]
#     6. chromEnd2 - chromosome end position for second interval [ 0-indexed, exclusive ]    
#     7. name - name of the feature
#     8. score - A score between 0 and 1000, can be used for filtering or visual cues
#     9. strand1 - strand for first interval | "+", "-", or "." (for don't know/don't care)
#     10. strand2 - strand for first interval | "+", "-", or "." (for don't know/don't care)
#     11. attributes -  raw string representing attribute list. converted and stored in 
#           attr_dict for convenience of use and filtering. should mirror format of GTF
@dataclass
class Bedpe:
    # raw information from object
    chrom1: str
    chromStart1: int # 0-indexed, beware interfacing with GTF objects
    chromEnd1: int # 0-indexed, inclusive, beware interfacing with GTF objects
    chrom2: str
    chromStart2: int # 0-indexed, beware interfacing with GTF objects
    chromEnd2: int # 0-indexed, inclusive, beware interfacing with GTF objects
    name: str
    score: float  # 0-1000 
    strand1: str  # '+', '-', or '.'
    strand2: str  # '+', '-', or '.'
    attributes: str  # raw string repr of the attributes, converted below

    # define post_init routines to create internal Bed6 objects for each
    # interval as well as an attribute dictionary shared between the two
    def __post_init__(self): # TODO
        self.bed1 = Bed6(
            self.chrom1, self.chromStart1, self.chromEnd1,
            self.name, self.score, self.strand1
        )
        self.bed2 = Bed6(
            self.chrom2, self.chromStart2, self.chromEnd2,
            self.name, self.score, self.strand2
        )

        # precompute first and second bed intervals in sort order to lessen
        # comparison burden in intersection and sorting operations
        self.first_bed = min([self.bed1,self.bed2])
        self.second_bed = max([self.bed1, self.bed2])

        # handle attributes field
        self.attr_dict = {}
        pre_proc_attr = self.attributes.split("; ")
        if pre_proc_attr != ['']:
            for pair in pre_proc_attr:
                k,v = [p.strip('"') for p in pair.strip(";").split()]
                if k not in self.attr_dict.keys():
                    self.attr_dict[k] = [check_and_convert_nums(v)]
                else:
                    self.attr_dict[k].append(check_and_convert_nums(v))

    # define custom representation when printing the Bedpe object
    def __repr__(self):
        spec = (
            f" {self.bed1} {self.bed2}"
            f" {self.attributes} "
        )
        return repr(f"Bedpe({spec})")

    # define a custom function for the equal to (==) comparator
    # based on identical interval information on matching coordinates
    def __eq__(self, other):
        comp = (((self.bed1 == other.bed1) and (self.bed2 == other.bed2)) or 
            ((self.bed1 == other.bed2) and (self.bed2 == other.bed1)))
        return comp
    
    # define a custom function for the (<) comparator. For Bedpe objects,
    # we define less than as the minimum of the two paired intervals 
    # compared to the minimum of the two paired intervals of the other object. 
    # if the minimum interval is identical, we compare the other interval from each 
    # This comparison only works for Bedpe to Bedpe objects.
    def __lt__(self, other):
        if lt(*[prep_chrom_comp(c) for c in [self.first_bed.chrom, other.first_bed.chrom]]): return True
        elif gt(*[prep_chrom_comp(c) for c in [self.first_bed.chrom, other.first_bed.chrom]]): return False
        else: # if the chromosomes names are equal by natural sort
            comp = ((self.first_bed < other.first_bed) or 
                    ((self.first_bed == other.first_bed) and 
                    (self.second_bed < other.second_bed)))
            return comp

    # define a custom function for the (>) comparator. For Bedpe objects,
    # we define greater than as the minimum of the two paired intervals 
    # compared to the minimum of the two paired intervals of the other object. 
    # This comparison only works for Bedpe to Bedpe objects.
    def __gt__(self, other):
        if lt(*[prep_chrom_comp(c) for c in [self.first_bed.chrom, other.first_bed.chrom]]): return False
        elif gt(*[prep_chrom_comp(c) for c in [self.first_bed.chrom, other.first_bed.chrom]]): return True
        else: # if the chromosomes names are equal by natural sort
            comp = ((self.first_bed > other.first_bed) or 
                    ((self.first_bed == other.first_bed) and 
                    (self.second_bed > other.second_bed)))
            return comp

    # define a custom function for the (<=) comparator. See __lt__ , __gt__ ,
    # and __eq__ as implemented above for specifics
    def __le__(self, other):
        return (self.__lt__(other) or self.__eq__(other))

    # define a custom function for the (>=) comparator. See __lt__ , __gt__ ,
    # and __eq__ as implemented above for specifics
    def __ge__(self,other):
        return (self.__gt__(other) or self.__eq__(other))



# File Name: __init__.py
# Created By: ZW
# Created On: 2022-12-19
# Purpose: submodule package init for gobjects.core

# module imports
# ----------------------------------------------------------------------------
from .bed import Bed6, Bed12
from .bedpe import Bedpe
from .gtf import Gtf
from .intervals import Interval
from .intervalmap import IntervalMap


# File Name: intersect.py
# Created By: ZW
# Created On: 2022-12-22
# Purpose: define functions for interval algebra ( much like those found
#  in bedtools ) specifically bedtools intersect-like functionality.

# module imports
# ----------------------------------------------------------------------------
from ..core.bed import Bed6, Bed12
from ..core.bedpe import Bedpe
from ..core.gtf import Gtf
from ..core.intervals import Interval
from ..core.intervalmap import IntervalNode, IntervalMap


# constants definitions
# ----------------------------------------------------------------------------
CONTIGUOUS = [Interval, Bed6, Bed12, Gtf, IntervalNode]
NON_CONTIGUOUS = [Bedpe]
ALL_GOBJECTS = CONTIGUOUS + NON_CONTIGUOUS

# function definitions
# ----------------------------------------------------------------------------

# define check_stranded() which takes an object and ensures it contains the
# self.strand attribute. If it does, the function returns None, if it doesn't
# the function raises an Attribute error.
def check_stranded(gobj):
    if not hasattr(gobj, "strand"):
        msg = (
            "Error! attemped strand-aware comparison"
            " for object with no strand information"
            )
        raise AttributeError(msg)
    else:
        return None

# define intersect() which takes two contiguous features, like Bed or GTF
# and returns the boolean result of the interval intersect. the function
# has a strand-aware option but is not checked internally for strand attrs
def intersect(gobjA, gobjB, strand_aware=False):
    if not all((type(gobjA) in CONTIGUOUS, type(gobjB) in CONTIGUOUS)):
        raise TypeError(
            "Cannot perform intersect() on non-contiguous features.\n"
            f"please ensure {gobjB} and/or {gobjB}. are one of {CONTIGUOUS}"
        )
    
    if strand_aware:  # check for strand information on strand aware items
        check_stranded(gobjA) 
        check_stranded(gobjB)
    
    if strand_aware and (gobjA.strand != gobjB.strand): return False
    else:  # regular intersect routine
        if gobjA.chrom != gobjB.chrom: return False
        else:
            return not (  # define NOT intersection conditions
                ((gobjA.zero_idx_end) < gobjB.zero_idx_start) or
                 (gobjA.zero_idx_start > (gobjB.zero_idx_end)))


# define function intersect_IntervalMap() which takes a query record
# and an IntervalMap and returns all of the records in the map
# which intersect the given query record.
def intersect_IntervalMap(query, interval_map):
    if not (type(query) in CONTIGUOUS):
        raise TypeError(
            "Cannot perform intersect() on non-contiguous features.\n"
            f"please ensure {query}. is one of {CONTIGUOUS}"
        )
    out_intersects = []  # output list of intersects
    bins = interval_map.get_map()[query.chrom]
    bin_intersects = [b for b in bins if intersect(query, b)]
    for b in bin_intersects:
        for rec in b.intervals:
            if intersect(query, rec): out_intersects.append(rec)
            elif rec > query: break
            else: continue
    return out_intersects


# File Name: __init__.py
# Created By: ZW
# Created On: 2022-01-09
# Purpose: submodule init for operations

# module imports
# ----------------------------------------------------------------------------
from .intersect import intersect, intersect_IntervalMap
from .writers import write_Bed6


# File Name: intervals.py
# Created By: ZW
# Created On: 2022-11-03
# Purpose: defines base class for BED style intervals commonly
#  used to describe linear features in the genome.


# module imports
# ----------------------------------------------------------------------------
from dataclasses import dataclass
from operator import lt, gt
from re import split as re_split

# function definitions
# ----------------------------------------------------------------------------

# function prep_chrom_compare() splits contig names that are a combo of
# strings and digits to use as sorting keys in sort functions and similar 
# processes. for example: 'chr1' -> ('chr', 1). this function can be 
# used in the sorted() function with the key=prep_chrom_comp argument
def prep_chrom_comp(chrom: str):
    def recode(substr: str):
        return int(substr) if substr.isdigit() else substr.lower()
    return [recode(sub) for sub in re_split('([0-9]+)',chrom)]

# class definitions
# ----------------------------------------------------------------------------

# class Interval() - base class for all BED-type genome interval objects. 
# * indexing of genomic intervals follow BED conventions. This means that
#   chromosome scaffolds begin at base 0. the the end of the interval is 
#   equal to 1 + the feature end. this represents half-open intervals of the type
#   [chromStart, chromEnd), and but importantly, they represent 
#   chrom:(chromStart+1)-chromEnd intervals in position notation, for example:
#   chr1:1-1000 is equivalent to the Interval "chr1 0 1000" or chr1: [0,1000)
#
#  * the class contains only four attributes:
#     1. chromosome (chrom) -required- 
#     2. chromosome start position (chromEnd) -required-
#     3. chromosome end position (chromEnd) -required- 
#     4. feature name (name) -required-
@dataclass(eq=False, order=False)
class Interval():
    chrom: str
    chromStart: int
    chromEnd: int
    name: str

    # define a post-init method to coerce ints and floats properly
    def __post_init__(self):
        self.chromStart = int(self.chromStart)
        self.chromEnd = int(self.chromEnd)
        self.zero_idx_start = self.chromStart  # always inclusive
        self.zero_idx_end = self.chromEnd - 1  # always inclusive

    # define a custom printout representation for the Interval
    def __repr__(self):
        spec = f" {self.chrom} {self.chromStart} {self.chromEnd} {self.name} "
        return repr(f"Interval({spec})")
    
    # define a custom function for the equal to (==) comparator
    # based on identical interval information on matching coordinates
    def __eq__(self, other):
        comp = ((self.chrom == other.chrom) and
                (self.chromStart == other.chromStart) and
                (self.chromEnd == other.chromEnd))
        return comp
    
    # define a custom function for the less than (<) comparator
    # based on interval algebra on matching chromosomes
    def __lt__(self, other):
        if lt(*[prep_chrom_comp(c) for c in [self.chrom, other.chrom]]): return True
        elif gt(*[prep_chrom_comp(c) for c in [self.chrom, other.chrom]]): return False
        else: # if the chromosomes names are equal by natural sort
            comp = ((self.chromStart < other.chromStart) or
                    ((self.chromStart == other.chromStart) and
                    (self.chromEnd < other.chromEnd)))
            return comp
    
    # define a custom function for the greater than (>) comparator
    # based on interval algebra on matching chromosomes
    def __gt__(self, other):
        if lt(*[prep_chrom_comp(c) for c in [self.chrom, other.chrom]]): return False
        elif gt(*[prep_chrom_comp(c) for c in [self.chrom, other.chrom]]): return True
        else: # if the chromosomes names are equal by natural sort
            comp = ((self.chromStart > other.chromStart) or 
                    ((self.chromEnd == other.chromEnd) and
                    (self.chromStart > other.chromStart)))
            return comp
    
    # define a custom function for the less than or equal to (<=) comparator
    # based on niterval algebra on matching chromosomes
    def __le__(self,other):
        return (self.__lt__(other) or self.__eq__(other))
    
    # define a custom function for the greater than or equal to (>=) comparator
    # based on interval algebra on matching chromosomes
    def __ge__(self,other):
        return (self.__gt__(other) or self.__eq__(other))
    




# File Name: writers.py
# Created By: ZW
# Created On: 2023-01-12
# Purpose: defines functions that write lists of gobjects to output text files
#  in the proper format. these writers in general are specific to a single 
#  record type and their corresponding file format.

# module imports
# ----------------------------------------------------------------------------
from gobjects.core import Bed6
from pathlib import Path


# function definitions
# ----------------------------------------------------------------------------

# define write_Bed6() which takes a list of bed6 records, and a file path and 
# writes the records out to a bed6 formatted file. optionally, you can append
# the records to a pre-existing file.
def write_Bed6(bed6_objects, filepath, fappend=False):
    # check that all objects in the passed list are Bed6
    check_types = [(type(b) == Bed6) for b in bed6_objects]
    if not all(check_types):
        raise TypeError("Error! not all record in passed list are Bed6\n")
    
    # open file pointer and ensure files to append exist
    fpath = Path(filepath).resolve()
    if not fappend: fobj = open(fpath, 'w')
    elif (fappend and fpath.exists()): fobj = open(fpath, 'a')
    else:
        msg = (
            f"Cannot find existing file {fpath} for append mode..\n"
            "plese ensure the file exists or switch to write mode"
        )
        raise FileNotFoundError(msg)
    
    # write records out sequentially
    print(f"Writing Bed6 records to {fpath}..")
    for record in bed6_objects:
        line = (
            f"{record.chrom}\t{record.chromStart}\t{record.chromEnd}\t"
            f"{record.name}\t{record.score}\t{record.strand}\n"
        )
        fobj.write(line)
    print(f"Done writing Bed6 records to {fpath}..")
    fobj.close()


# File Name: setup.py
# Created By: ZW
# Created On: 2022-11-02
# Puropse: defines package information and
# requirements for the gobjects package

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gobjects",
    version="0.0.6",
    author="Zach Weber",
    author_email="zach.weber.813@gmail.com",
    description="Python3 genetics objects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zwebbs/gobjects",
    project_urls={
        "Bug Tracker": "https://github.com/zwebbs/gobjects/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        'Operating System :: POSIX',
        'Operating System :: MacOS',
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.7",
    install_requires=[],
    extras_require={
        'dev': ['build==0.9.0', 'twine==4.0.1']
    }
)



# File Name: gtf.py
# Created By: ZW
# Created On: 2022-12-20
# Purpose: defines class for gtf style objects (identical to the GFF2 spec),
#  as defined by standards at https://genome.ucsc.edu/FAQ/FAQformat.html#format3

# module imports
# ----------------------------------------------------------------------------
from dataclasses import dataclass
from operator import lt, gt
from re import split as re_split
from re import match
from typing import Union

# function definitions
# ----------------------------------------------------------------------------

# function prep_chrom_compare() splits contig names that are a combo of
# strings and digits to use as sorting keys in sort functions and similar 
# processes. for example: 'chr1' -> ('chr', 1). this function can be 
# used in the sorted() function with the key=prep_chrom_comp argument
def prep_chrom_comp(chrom: str):
    def recode(substr: str):
        return int(substr) if substr.isdigit() else substr.lower()
    return [recode(sub) for sub in re_split('([0-9]+)',chrom)]

# function check_and_convert_nums() takes a string and converts it to 
# a floating point python number if it looks like an integer or float,
# otherwise, it returns the string back unchanged
def check_and_convert_nums(string):
    pattern = "^[+-]?((\d+(\.\d+)?)|(\.\d+))$"
    mth = match(pattern, string)
    return (float(mth.string) if mth is not None else string)

# class definitions
# ----------------------------------------------------------------------------

# class Gtf() - base class for all GTF-type gene-like objects. 
# * indexing of genomic intervals follow GTF conventions. This means that
#   chromosome scaffolds begin at base 1. (This differs from bed-syle so care
#   must be taken when intersecting the two object types) the end of the interval
#   is inclusive [chromStart, chromEnd], creating slightly different intersection
#   rules than bed-style objects. 
#
# * The Gtf class contains nine attributes: 
#     1. chrom - chromosome of the feature
#     2. source - The program that generated this feature
#     3. feature - The name of the type of feature e.g. (exon, transcript, CDS, etc. )
#     4. chromStart - chromosome start position [ 1-indexed, inclusive ]
#     5. chromEnd - chromosome end position [ 1-indexed, inclusive ]
#     6. score - A score between 0 and 1000, can be used for filtering or visual cues
#     7. strand - strand options include "+", "-", or "." (for don't know/don't care)
#     8. frame - for exons and similar, frame should be a number between 0-2
#     9. attributes -  raw string representing attribute list. converted and stored in 
#           attr_dict for convenience of use and filtering.
@dataclass
class Gtf:
    chrom: str
    source: str  # e.g. HAVANA, ncbi, etc
    feature: str 
    chromStart: int # 1-indexed, beware interfacing with BED objects
    chromEnd: int # 1-indexed, inclusive, beware interfacing with BED objects
    score: float  # 0-1000 
    strand: str  # '+', '-', or '.'
    frame: Union[str,int]  # 0-2 or '.'
    attributes: str  # raw string repr of the attributes, converted below
    
    # define post_init routines to create attribute dictionary
    def __post_init__(self): # TODO
        self.chromStart = int(self.chromStart)
        self.chromEnd = int(self.chromEnd)
        self.score = check_and_convert_nums(str(self.score))
        self.zero_idx_start = self.chromStart - 1  # always inclusive
        self.zero_idx_end = self.chromEnd - 1  #always inclusive

        # handle attributes field
        self.attr_dict = {}
        pre_proc_attr = self.attributes.split("; ")
        if pre_proc_attr != ['']:
            for pair in pre_proc_attr:
                k,v = [p.strip('"') for p in pair.strip(";").split()]
                if k not in self.attr_dict.keys():
                    self.attr_dict[k] = [check_and_convert_nums(v)]
                else:
                    self.attr_dict[k].append(check_and_convert_nums(v))

    # define custom representation when printing the GTF object
    def __repr__(self):
        spec = (
            f" {self.chrom} {self.source} {self.feature}"
            f" {self.chromStart} {self.chromEnd} {self.score}"
            f" {self.strand} {self.frame} {self.attributes} "
        )
        return repr(f"Gtf({spec})")

        # define a custom function for the equal to (==) comparator
    # based on identical interval information on matching coordinates
    def __eq__(self, other):
        comp = ((self.chrom == other.chrom) and
                (self.chromStart == other.chromStart) and
                (self.chromEnd == other.chromEnd))
        return comp
    
    # define a custom function for the less than (<) comparator
    # based on interval algebra on matching chromosomes
    def __lt__(self, other):
        if lt(*[prep_chrom_comp(c) for c in [self.chrom, other.chrom]]): return True
        elif gt(*[prep_chrom_comp(c) for c in [self.chrom, other.chrom]]): return False
        else: # if the chromosomes names are equal by natural sort
            comp = ((self.chromStart < other.chromStart) or
                    ((self.chromStart == other.chromStart) and
                    (self.chromEnd < other.chromEnd)))
            return comp
    
    # define a custom function for the greater than (>) comparator
    # based on interval algebra on matching chromosomes
    def __gt__(self, other):
        if lt(*[prep_chrom_comp(c) for c in [self.chrom, other.chrom]]): return False
        elif gt(*[prep_chrom_comp(c) for c in [self.chrom, other.chrom]]): return True
        else: # if the chromosomes names are equal by natural sort
            comp = ((self.chromStart > other.chromStart) or 
                    ((self.chromEnd == other.chromEnd) and
                    (self.chromStart > other.chromStart)))
            return comp
    
    # define a custom function for the less than or equal to (<=) comparator
    # based on niterval algebra on matching chromosomes
    def __le__(self,other):
        return (self.__lt__(other) or self.__eq__(other))
    
    # define a custom function for the greater than or equal to (>=) comparator
    # based on interval algebra on matching chromosomes
    def __ge__(self,other):
        return (self.__gt__(other) or self.__eq__(other))
